# THIS FILE WAS AUTOMATICALLY GENERATED BY "generate_dicts_from_data_json.py" DO NOT CHANGE MANUALLY!
# ANY CHANGE WILL BE OVERWRITTEN

from sc2.ids.ability_id import AbilityId
from sc2.ids.unit_typeid import UnitTypeId

# from sc2.ids.buff_id import BuffId
# from sc2.ids.effect_id import EffectId


UNIT_ABILITIES: dict[UnitTypeId, set[AbilityId]] = {
    UnitTypeId.ADEPT: {
        AbilityId.ADEPTPHASESHIFT_ADEPTPHASESHIFT,
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.ADEPTPHASESHIFT: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.CANCEL_ADEPTSHADEPHASESHIFT,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.ARBITERMP: {
        AbilityId.ARBITERMPRECALL_ARBITERMPRECALL,
        AbilityId.ARBITERMPSTASISFIELD_ARBITERMPSTASISFIELD,
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.ARCHON: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.ARMORY: {
        AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL1,
        AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL2,
        AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL3,
        AbilityId.ARMORYRESEARCH_TERRANVEHICLEANDSHIPPLATINGLEVEL1,
        AbilityId.ARMORYRESEARCH_TERRANVEHICLEANDSHIPPLATINGLEVEL2,
        AbilityId.ARMORYRESEARCH_TERRANVEHICLEANDSHIPPLATINGLEVEL3,
        AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL1,
        AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL2,
        AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL3,
    },
    UnitTypeId.AUTOTURRET: {AbilityId.ATTACK_ATTACK, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.BANELING: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BEHAVIOR_BUILDINGATTACKON,
        AbilityId.BURROWDOWN_BANELING,
        AbilityId.EXPLODE_EXPLODE,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.BANELINGBURROWED: {AbilityId.BURROWUP_BANELING, AbilityId.EXPLODE_EXPLODE},
    UnitTypeId.BANELINGCOCOON: {AbilityId.RALLY_BUILDING, AbilityId.SMART},
    UnitTypeId.BANELINGNEST: {AbilityId.RESEARCH_CENTRIFUGALHOOKS},
    UnitTypeId.BANSHEE: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BEHAVIOR_CLOAKON_BANSHEE,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.BARRACKS: {
        AbilityId.BARRACKSTRAIN_GHOST,
        AbilityId.BARRACKSTRAIN_MARAUDER,
        AbilityId.BARRACKSTRAIN_MARINE,
        AbilityId.BARRACKSTRAIN_REAPER,
        AbilityId.BUILD_REACTOR_BARRACKS,
        AbilityId.BUILD_TECHLAB_BARRACKS,
        AbilityId.LIFT_BARRACKS,
        AbilityId.RALLY_BUILDING,
        AbilityId.SMART,
    },
    UnitTypeId.BARRACKSFLYING: {
        AbilityId.BUILD_REACTOR_BARRACKS,
        AbilityId.BUILD_TECHLAB_BARRACKS,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.LAND_BARRACKS,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.BARRACKSTECHLAB: {
        AbilityId.BARRACKSTECHLABRESEARCH_STIMPACK,
        AbilityId.RESEARCH_COMBATSHIELD,
        AbilityId.RESEARCH_CONCUSSIVESHELLS,
    },
    UnitTypeId.BATTLECRUISER: {
        AbilityId.ATTACK_BATTLECRUISER,
        AbilityId.EFFECT_TACTICALJUMP,
        AbilityId.HOLDPOSITION_BATTLECRUISER,
        AbilityId.MOVE_BATTLECRUISER,
        AbilityId.PATROL_BATTLECRUISER,
        AbilityId.SMART,
        AbilityId.STOP_BATTLECRUISER,
        AbilityId.YAMATO_YAMATOGUN,
    },
    UnitTypeId.BROODLING: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.BROODLORD: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.BUNKER: {
        AbilityId.LOAD_BUNKER,
        AbilityId.RALLY_BUILDING,
        AbilityId.SALVAGEEFFECT_SALVAGE,
        AbilityId.SMART,
    },
    UnitTypeId.BYPASSARMORDRONE: {AbilityId.ATTACK_ATTACK, AbilityId.MOVE_MOVE, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.CARRIER: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BUILD_INTERCEPTORS,
        AbilityId.CANCEL_HANGARQUEUE5,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.CHANGELING: {
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.CHANGELINGMARINE: {
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.CHANGELINGMARINESHIELD: {
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.CHANGELINGZEALOT: {
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.CHANGELINGZERGLING: {
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.CHANGELINGZERGLINGWINGS: {
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.COLOSSUS: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.COMMANDCENTER: {
        AbilityId.COMMANDCENTERTRAIN_SCV,
        AbilityId.LIFT_COMMANDCENTER,
        AbilityId.LOADALL_COMMANDCENTER,
        AbilityId.RALLY_COMMANDCENTER,
        AbilityId.SMART,
        AbilityId.UPGRADETOORBITAL_ORBITALCOMMAND,
        AbilityId.UPGRADETOPLANETARYFORTRESS_PLANETARYFORTRESS,
    },
    UnitTypeId.COMMANDCENTERFLYING: {
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.LAND_COMMANDCENTER,
        AbilityId.LOADALL_COMMANDCENTER,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.CORRUPTOR: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.CAUSTICSPRAY_CAUSTICSPRAY,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPHTOBROODLORD_BROODLORD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.CORSAIRMP: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.CORSAIRMPDISRUPTIONWEB_CORSAIRMPDISRUPTIONWEB,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.CREEPTUMORBURROWED: {AbilityId.BUILD_CREEPTUMOR, AbilityId.BUILD_CREEPTUMOR_TUMOR, AbilityId.SMART},
    UnitTypeId.CYBERNETICSCORE: {
        AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRARMORLEVEL1,
        AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRARMORLEVEL2,
        AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRARMORLEVEL3,
        AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL1,
        AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL2,
        AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL3,
        AbilityId.RESEARCH_WARPGATE,
    },
    UnitTypeId.CYCLONE: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.LOCKON_LOCKON,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.DARKSHRINE: {AbilityId.RESEARCH_SHADOWSTRIKE},
    UnitTypeId.DARKTEMPLAR: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.EFFECT_SHADOWSTRIDE,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPH_ARCHON,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.DEFILERMP: {
        AbilityId.DEFILERMPBURROW_BURROWDOWN,
        AbilityId.DEFILERMPCONSUME_DEFILERMPCONSUME,
        AbilityId.DEFILERMPDARKSWARM_DEFILERMPDARKSWARM,
        AbilityId.DEFILERMPPLAGUE_DEFILERMPPLAGUE,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.DEFILERMPBURROWED: {AbilityId.DEFILERMPUNBURROW_BURROWUP},
    UnitTypeId.DEVOURERMP: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.DISRUPTOR: {
        AbilityId.EFFECT_PURIFICATIONNOVA,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.DISRUPTORPHASED: {
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.DRONE: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BUILD_LURKERDEN,
        AbilityId.BURROWDOWN_DRONE,
        AbilityId.EFFECT_SPRAY_ZERG,
        AbilityId.HARVEST_GATHER_DRONE,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
        AbilityId.ZERGBUILD_BANELINGNEST,
        AbilityId.ZERGBUILD_EVOLUTIONCHAMBER,
        AbilityId.ZERGBUILD_EXTRACTOR,
        AbilityId.ZERGBUILD_HATCHERY,
        AbilityId.ZERGBUILD_HYDRALISKDEN,
        AbilityId.ZERGBUILD_INFESTATIONPIT,
        AbilityId.ZERGBUILD_NYDUSNETWORK,
        AbilityId.ZERGBUILD_ROACHWARREN,
        AbilityId.ZERGBUILD_SPAWNINGPOOL,
        AbilityId.ZERGBUILD_SPINECRAWLER,
        AbilityId.ZERGBUILD_SPIRE,
        AbilityId.ZERGBUILD_SPORECRAWLER,
        AbilityId.ZERGBUILD_ULTRALISKCAVERN,
    },
    UnitTypeId.DRONEBURROWED: {AbilityId.BURROWUP_DRONE},
    UnitTypeId.EGG: {AbilityId.RALLY_BUILDING, AbilityId.SMART},
    UnitTypeId.ELSECARO_COLONIST_HUT: {AbilityId.RALLY_BUILDING, AbilityId.SMART},
    UnitTypeId.ENGINEERINGBAY: {
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL1,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL2,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL3,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL1,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL2,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL3,
        AbilityId.RESEARCH_HISECAUTOTRACKING,
        AbilityId.RESEARCH_TERRANSTRUCTUREARMORUPGRADE,
    },
    UnitTypeId.EVOLUTIONCHAMBER: {
        AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL1,
        AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL2,
        AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL3,
        AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL1,
        AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL2,
        AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL3,
        AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL1,
        AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL2,
        AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL3,
    },
    UnitTypeId.FACTORY: {
        AbilityId.BUILD_REACTOR_FACTORY,
        AbilityId.BUILD_TECHLAB_FACTORY,
        AbilityId.FACTORYTRAIN_HELLION,
        AbilityId.FACTORYTRAIN_SIEGETANK,
        AbilityId.FACTORYTRAIN_THOR,
        AbilityId.FACTORYTRAIN_WIDOWMINE,
        AbilityId.LIFT_FACTORY,
        AbilityId.RALLY_BUILDING,
        AbilityId.SMART,
        AbilityId.TRAIN_CYCLONE,
        AbilityId.TRAIN_HELLBAT,
    },
    UnitTypeId.FACTORYFLYING: {
        AbilityId.BUILD_REACTOR_FACTORY,
        AbilityId.BUILD_TECHLAB_FACTORY,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.LAND_FACTORY,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.FACTORYTECHLAB: {
        AbilityId.RESEARCH_CYCLONELOCKONDAMAGE,
        AbilityId.RESEARCH_DRILLINGCLAWS,
        AbilityId.RESEARCH_INFERNALPREIGNITER,
        AbilityId.RESEARCH_SMARTSERVOS,
    },
    UnitTypeId.FLEETBEACON: {
        AbilityId.FLEETBEACONRESEARCH_RESEARCHVOIDRAYSPEEDUPGRADE,
        AbilityId.FLEETBEACONRESEARCH_TEMPESTRESEARCHGROUNDATTACKUPGRADE,
        AbilityId.RESEARCH_PHOENIXANIONPULSECRYSTALS,
    },
    UnitTypeId.FORGE: {
        AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL1,
        AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL2,
        AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL3,
        AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL1,
        AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL2,
        AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL3,
        AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL1,
        AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL2,
        AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL3,
    },
    UnitTypeId.FUSIONCORE: {
        AbilityId.FUSIONCORERESEARCH_RESEARCHBALLISTICRANGE,
        AbilityId.FUSIONCORERESEARCH_RESEARCHMEDIVACENERGYUPGRADE,
        AbilityId.RESEARCH_BATTLECRUISERWEAPONREFIT,
    },
    UnitTypeId.GATEWAY: {
        AbilityId.GATEWAYTRAIN_DARKTEMPLAR,
        AbilityId.GATEWAYTRAIN_HIGHTEMPLAR,
        AbilityId.GATEWAYTRAIN_SENTRY,
        AbilityId.GATEWAYTRAIN_STALKER,
        AbilityId.GATEWAYTRAIN_ZEALOT,
        AbilityId.MORPH_WARPGATE,
        AbilityId.RALLY_BUILDING,
        AbilityId.SMART,
        AbilityId.TRAIN_ADEPT,
    },
    UnitTypeId.GHOST: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BEHAVIOR_CLOAKON_GHOST,
        AbilityId.BEHAVIOR_HOLDFIREON_GHOST,
        AbilityId.EFFECT_GHOSTSNIPE,
        AbilityId.EMP_EMP,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.GHOSTACADEMY: {AbilityId.BUILD_NUKE, AbilityId.RESEARCH_PERSONALCLOAKING},
    UnitTypeId.GHOSTNOVA: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BEHAVIOR_CLOAKON_GHOST,
        AbilityId.BEHAVIOR_HOLDFIREON_GHOST,
        AbilityId.EFFECT_GHOSTSNIPE,
        AbilityId.EMP_EMP,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.GREATERSPIRE: {
        AbilityId.RESEARCH_ZERGFLYERARMORLEVEL1,
        AbilityId.RESEARCH_ZERGFLYERARMORLEVEL2,
        AbilityId.RESEARCH_ZERGFLYERARMORLEVEL3,
        AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL1,
        AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL2,
        AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL3,
    },
    UnitTypeId.GUARDIANMP: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.HATCHERY: {
        AbilityId.RALLY_HATCHERY_UNITS,
        AbilityId.RALLY_HATCHERY_WORKERS,
        AbilityId.RESEARCH_BURROW,
        AbilityId.RESEARCH_PNEUMATIZEDCARAPACE,
        AbilityId.SMART,
        AbilityId.TRAINQUEEN_QUEEN,
        AbilityId.UPGRADETOLAIR_LAIR,
    },
    UnitTypeId.HELLION: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPH_HELLBAT,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.HELLIONTANK: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPH_HELLION,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.HERC: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.HERCPLACEMENT: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.HIGHTEMPLAR: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.FEEDBACK_FEEDBACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPH_ARCHON,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.PSISTORM_PSISTORM,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.HIVE: {
        AbilityId.RALLY_HATCHERY_UNITS,
        AbilityId.RALLY_HATCHERY_WORKERS,
        AbilityId.RESEARCH_BURROW,
        AbilityId.RESEARCH_PNEUMATIZEDCARAPACE,
        AbilityId.SMART,
        AbilityId.TRAINQUEEN_QUEEN,
    },
    UnitTypeId.HYDRALISK: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BURROWDOWN_HYDRALISK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.HYDRALISKFRENZY_HYDRALISKFRENZY,
        AbilityId.MORPH_LURKER,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.HYDRALISKBURROWED: {AbilityId.BURROWUP_HYDRALISK},
    UnitTypeId.HYDRALISKDEN: {
        AbilityId.HYDRALISKDENRESEARCH_RESEARCHFRENZY,
        AbilityId.RESEARCH_GROOVEDSPINES,
        AbilityId.RESEARCH_MUSCULARAUGMENTS,
    },
    UnitTypeId.IMMORTAL: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.INFESTATIONPIT: {AbilityId.RESEARCH_NEURALPARASITE},
    UnitTypeId.INFESTOR: {
        AbilityId.AMORPHOUSARMORCLOUD_AMORPHOUSARMORCLOUD,
        AbilityId.BURROWDOWN_INFESTOR,
        AbilityId.BURROWDOWN_INFESTORTERRAN,
        AbilityId.FUNGALGROWTH_FUNGALGROWTH,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.NEURALPARASITE_NEURALPARASITE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.INFESTORBURROWED: {
        AbilityId.BURROWUP_INFESTOR,
        AbilityId.BURROWUP_INFESTORTERRAN,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.NEURALPARASITE_NEURALPARASITE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.INFESTORTERRAN: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BURROWDOWN_INFESTORTERRAN,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.INFESTORTERRANBURROWED: {AbilityId.BURROWUP_INFESTORTERRAN},
    UnitTypeId.INTERCEPTOR: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.LAIR: {
        AbilityId.RALLY_HATCHERY_UNITS,
        AbilityId.RALLY_HATCHERY_WORKERS,
        AbilityId.RESEARCH_BURROW,
        AbilityId.RESEARCH_PNEUMATIZEDCARAPACE,
        AbilityId.SMART,
        AbilityId.TRAINQUEEN_QUEEN,
        AbilityId.UPGRADETOHIVE_HIVE,
    },
    UnitTypeId.LARVA: {
        AbilityId.LARVATRAIN_CORRUPTOR,
        AbilityId.LARVATRAIN_DRONE,
        AbilityId.LARVATRAIN_HYDRALISK,
        AbilityId.LARVATRAIN_INFESTOR,
        AbilityId.LARVATRAIN_MUTALISK,
        AbilityId.LARVATRAIN_OVERLORD,
        AbilityId.LARVATRAIN_ROACH,
        AbilityId.LARVATRAIN_ULTRALISK,
        AbilityId.LARVATRAIN_VIPER,
        AbilityId.LARVATRAIN_ZERGLING,
        AbilityId.TRAIN_SWARMHOST,
    },
    UnitTypeId.LIBERATOR: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPH_LIBERATORAGMODE,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.LIBERATORAG: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.MORPH_LIBERATORAAMODE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.LOCUSTMP: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.LOCUSTMPFLYING: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.EFFECT_LOCUSTSWOOP,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.LURKERDENMP: {AbilityId.LURKERDENRESEARCH_RESEARCHLURKERRANGE, AbilityId.RESEARCH_ADAPTIVETALONS},
    UnitTypeId.LURKERMP: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BURROWDOWN_LURKER,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.LURKERMPBURROWED: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BEHAVIOR_HOLDFIREON_LURKER,
        AbilityId.BURROWUP_LURKER,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.LURKERMPEGG: {AbilityId.RALLY_BUILDING, AbilityId.SMART},
    UnitTypeId.MARAUDER: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.EFFECT_STIM_MARAUDER,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.MARINE: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.EFFECT_STIM_MARINE,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.MEDIVAC: {
        AbilityId.EFFECT_MEDIVACIGNITEAFTERBURNERS,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.LOAD_MEDIVAC,
        AbilityId.MEDIVACHEAL_HEAL,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.MISSILETURRET: {AbilityId.ATTACK_ATTACK, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.MOTHERSHIP: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.EFFECT_MASSRECALL_STRATEGICRECALL,
        AbilityId.EFFECT_TIMEWARP,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOTHERSHIPCLOAK_ORACLECLOAKFIELD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.MOTHERSHIPCORE: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.EFFECT_MASSRECALL_MOTHERSHIPCORE,
        AbilityId.EFFECT_PHOTONOVERCHARGE,
        AbilityId.EFFECT_TIMEWARP,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPH_MOTHERSHIP,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.MULE: {
        AbilityId.EFFECT_REPAIR_MULE,
        AbilityId.HARVEST_GATHER_MULE,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.MUTALISK: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.NEXUS: {
        AbilityId.EFFECT_CHRONOBOOSTENERGYCOST,
        AbilityId.EFFECT_MASSRECALL_NEXUS,
        AbilityId.ENERGYRECHARGE_ENERGYRECHARGE,
        AbilityId.NEXUSTRAINMOTHERSHIP_MOTHERSHIP,
        AbilityId.NEXUSTRAIN_PROBE,
        AbilityId.RALLY_NEXUS,
        AbilityId.SMART,
    },
    UnitTypeId.NYDUSCANAL: {AbilityId.LOAD_NYDUSWORM, AbilityId.RALLY_BUILDING, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.NYDUSCANALATTACKER: {AbilityId.ATTACK_ATTACK, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.NYDUSCANALCREEPER: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.DIGESTERCREEPSPRAY_DIGESTERCREEPSPRAY,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.NYDUSNETWORK: {
        AbilityId.BUILD_NYDUSWORM,
        AbilityId.LOAD_NYDUSNETWORK,
        AbilityId.RALLY_BUILDING,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.OBSERVER: {
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPH_SURVEILLANCEMODE,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.OBSERVERSIEGEMODE: {AbilityId.MORPH_OBSERVERMODE, AbilityId.STOP_STOP},
    UnitTypeId.ORACLE: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BEHAVIOR_PULSARBEAMON,
        AbilityId.BUILD_STASISTRAP,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.ORACLEREVELATION_ORACLEREVELATION,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.ORBITALCOMMAND: {
        AbilityId.CALLDOWNMULE_CALLDOWNMULE,
        AbilityId.COMMANDCENTERTRAIN_SCV,
        AbilityId.LIFT_ORBITALCOMMAND,
        AbilityId.RALLY_COMMANDCENTER,
        AbilityId.SCANNERSWEEP_SCAN,
        AbilityId.SMART,
        AbilityId.SUPPLYDROP_SUPPLYDROP,
    },
    UnitTypeId.ORBITALCOMMANDFLYING: {
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.LAND_ORBITALCOMMAND,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.OVERLORD: {
        AbilityId.BEHAVIOR_GENERATECREEPON,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPH_OVERLORDTRANSPORT,
        AbilityId.MORPH_OVERSEER,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.OVERLORDTRANSPORT: {
        AbilityId.BEHAVIOR_GENERATECREEPON,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.LOAD_OVERLORD,
        AbilityId.MORPH_OVERSEER,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.OVERSEER: {
        AbilityId.CONTAMINATE_CONTAMINATE,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPH_OVERSIGHTMODE,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.SPAWNCHANGELING_SPAWNCHANGELING,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.OVERSEERSIEGEMODE: {
        AbilityId.CONTAMINATE_CONTAMINATE,
        AbilityId.MORPH_OVERSEERMODE,
        AbilityId.SMART,
        AbilityId.SPAWNCHANGELING_SPAWNCHANGELING,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.PHOENIX: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.GRAVITONBEAM_GRAVITONBEAM,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.PHOTONCANNON: {AbilityId.ATTACK_ATTACK, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.PLANETARYFORTRESS: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.COMMANDCENTERTRAIN_SCV,
        AbilityId.LOADALL_COMMANDCENTER,
        AbilityId.RALLY_COMMANDCENTER,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.PROBE: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BUILD_SHIELDBATTERY,
        AbilityId.EFFECT_SPRAY_PROTOSS,
        AbilityId.HARVEST_GATHER_PROBE,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.PROTOSSBUILD_ASSIMILATOR,
        AbilityId.PROTOSSBUILD_CYBERNETICSCORE,
        AbilityId.PROTOSSBUILD_DARKSHRINE,
        AbilityId.PROTOSSBUILD_FLEETBEACON,
        AbilityId.PROTOSSBUILD_FORGE,
        AbilityId.PROTOSSBUILD_GATEWAY,
        AbilityId.PROTOSSBUILD_NEXUS,
        AbilityId.PROTOSSBUILD_PHOTONCANNON,
        AbilityId.PROTOSSBUILD_PYLON,
        AbilityId.PROTOSSBUILD_ROBOTICSBAY,
        AbilityId.PROTOSSBUILD_ROBOTICSFACILITY,
        AbilityId.PROTOSSBUILD_STARGATE,
        AbilityId.PROTOSSBUILD_TEMPLARARCHIVE,
        AbilityId.PROTOSSBUILD_TWILIGHTCOUNCIL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.QUEEN: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BUILD_CREEPTUMOR,
        AbilityId.BUILD_CREEPTUMOR_QUEEN,
        AbilityId.BURROWDOWN_QUEEN,
        AbilityId.EFFECT_INJECTLARVA,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
        AbilityId.TRANSFUSION_TRANSFUSION,
    },
    UnitTypeId.QUEENBURROWED: {AbilityId.BURROWUP_QUEEN},
    UnitTypeId.QUEENMP: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.QUEENMPENSNARE_QUEENMPENSNARE,
        AbilityId.QUEENMPINFESTCOMMANDCENTER_QUEENMPINFESTCOMMANDCENTER,
        AbilityId.QUEENMPSPAWNBROODLINGS_QUEENMPSPAWNBROODLINGS,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.RAVAGER: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BURROWDOWN_RAVAGER,
        AbilityId.EFFECT_CORROSIVEBILE,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.RAVAGERBURROWED: {AbilityId.BURROWUP_RAVAGER},
    UnitTypeId.RAVAGERCOCOON: {AbilityId.RALLY_BUILDING, AbilityId.SMART},
    UnitTypeId.RAVEN: {
        AbilityId.BUILDAUTOTURRET_AUTOTURRET,
        AbilityId.EFFECT_ANTIARMORMISSILE,
        AbilityId.EFFECT_INTERFERENCEMATRIX,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.RAVENREPAIRDRONE: {AbilityId.EFFECT_REPAIR_REPAIRDRONE, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.REAPER: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.KD8CHARGE_KD8CHARGE,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.REPLICANT: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.ROACH: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BURROWDOWN_ROACH,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPHTORAVAGER_RAVAGER,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.ROACHBURROWED: {
        AbilityId.BURROWUP_ROACH,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.ROACHWARREN: {AbilityId.RESEARCH_GLIALREGENERATION, AbilityId.RESEARCH_TUNNELINGCLAWS},
    UnitTypeId.ROBOTICSBAY: {
        AbilityId.RESEARCH_EXTENDEDTHERMALLANCE,
        AbilityId.RESEARCH_GRAVITICBOOSTER,
        AbilityId.RESEARCH_GRAVITICDRIVE,
    },
    UnitTypeId.ROBOTICSFACILITY: {
        AbilityId.RALLY_BUILDING,
        AbilityId.ROBOTICSFACILITYTRAIN_COLOSSUS,
        AbilityId.ROBOTICSFACILITYTRAIN_IMMORTAL,
        AbilityId.ROBOTICSFACILITYTRAIN_OBSERVER,
        AbilityId.ROBOTICSFACILITYTRAIN_WARPPRISM,
        AbilityId.SMART,
        AbilityId.TRAIN_DISRUPTOR,
    },
    UnitTypeId.SCOURGEMP: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.SCOUTMP: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.SCV: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.EFFECT_REPAIR_SCV,
        AbilityId.EFFECT_SPRAY_TERRAN,
        AbilityId.HARVEST_GATHER_SCV,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
        AbilityId.TERRANBUILD_ARMORY,
        AbilityId.TERRANBUILD_BARRACKS,
        AbilityId.TERRANBUILD_BUNKER,
        AbilityId.TERRANBUILD_COMMANDCENTER,
        AbilityId.TERRANBUILD_ENGINEERINGBAY,
        AbilityId.TERRANBUILD_FACTORY,
        AbilityId.TERRANBUILD_FUSIONCORE,
        AbilityId.TERRANBUILD_GHOSTACADEMY,
        AbilityId.TERRANBUILD_MISSILETURRET,
        AbilityId.TERRANBUILD_REFINERY,
        AbilityId.TERRANBUILD_SENSORTOWER,
        AbilityId.TERRANBUILD_STARPORT,
        AbilityId.TERRANBUILD_SUPPLYDEPOT,
    },
    UnitTypeId.SENSORTOWER: {AbilityId.SALVAGEEFFECT_SALVAGE},
    UnitTypeId.SENTRY: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.FORCEFIELD_FORCEFIELD,
        AbilityId.GUARDIANSHIELD_GUARDIANSHIELD,
        AbilityId.HALLUCINATION_ADEPT,
        AbilityId.HALLUCINATION_ARCHON,
        AbilityId.HALLUCINATION_COLOSSUS,
        AbilityId.HALLUCINATION_DISRUPTOR,
        AbilityId.HALLUCINATION_HIGHTEMPLAR,
        AbilityId.HALLUCINATION_IMMORTAL,
        AbilityId.HALLUCINATION_ORACLE,
        AbilityId.HALLUCINATION_PHOENIX,
        AbilityId.HALLUCINATION_PROBE,
        AbilityId.HALLUCINATION_STALKER,
        AbilityId.HALLUCINATION_VOIDRAY,
        AbilityId.HALLUCINATION_WARPPRISM,
        AbilityId.HALLUCINATION_ZEALOT,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.SHIELDBATTERY: {AbilityId.SHIELDBATTERYRECHARGEEX5_SHIELDBATTERYRECHARGE, AbilityId.SMART},
    UnitTypeId.SIEGETANK: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SIEGEMODE_SIEGEMODE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.SIEGETANKSIEGED: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
        AbilityId.UNSIEGE_UNSIEGE,
    },
    UnitTypeId.SPAWNINGPOOL: {AbilityId.RESEARCH_ZERGLINGADRENALGLANDS, AbilityId.RESEARCH_ZERGLINGMETABOLICBOOST},
    UnitTypeId.SPINECRAWLER: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.SMART,
        AbilityId.SPINECRAWLERUPROOT_SPINECRAWLERUPROOT,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.SPINECRAWLERUPROOTED: {
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.SPINECRAWLERROOT_SPINECRAWLERROOT,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.SPIRE: {
        AbilityId.RESEARCH_ZERGFLYERARMORLEVEL1,
        AbilityId.RESEARCH_ZERGFLYERARMORLEVEL2,
        AbilityId.RESEARCH_ZERGFLYERARMORLEVEL3,
        AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL1,
        AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL2,
        AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL3,
        AbilityId.UPGRADETOGREATERSPIRE_GREATERSPIRE,
    },
    UnitTypeId.SPORECRAWLER: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.SMART,
        AbilityId.SPORECRAWLERUPROOT_SPORECRAWLERUPROOT,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.SPORECRAWLERUPROOTED: {
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.SPORECRAWLERROOT_SPORECRAWLERROOT,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.STALKER: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.EFFECT_BLINK_STALKER,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.STARGATE: {
        AbilityId.RALLY_BUILDING,
        AbilityId.SMART,
        AbilityId.STARGATETRAIN_CARRIER,
        AbilityId.STARGATETRAIN_ORACLE,
        AbilityId.STARGATETRAIN_PHOENIX,
        AbilityId.STARGATETRAIN_TEMPEST,
        AbilityId.STARGATETRAIN_VOIDRAY,
    },
    UnitTypeId.STARPORT: {
        AbilityId.BUILD_REACTOR_STARPORT,
        AbilityId.BUILD_TECHLAB_STARPORT,
        AbilityId.LIFT_STARPORT,
        AbilityId.RALLY_BUILDING,
        AbilityId.SMART,
        AbilityId.STARPORTTRAIN_BANSHEE,
        AbilityId.STARPORTTRAIN_BATTLECRUISER,
        AbilityId.STARPORTTRAIN_LIBERATOR,
        AbilityId.STARPORTTRAIN_MEDIVAC,
        AbilityId.STARPORTTRAIN_RAVEN,
        AbilityId.STARPORTTRAIN_VIKINGFIGHTER,
    },
    UnitTypeId.STARPORTFLYING: {
        AbilityId.BUILD_REACTOR_STARPORT,
        AbilityId.BUILD_TECHLAB_STARPORT,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.LAND_STARPORT,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.STARPORTTECHLAB: {
        AbilityId.RESEARCH_BANSHEECLOAKINGFIELD,
        AbilityId.RESEARCH_BANSHEEHYPERFLIGHTROTORS,
        AbilityId.STARPORTTECHLABRESEARCH_RESEARCHRAVENINTERFERENCEMATRIX,
    },
    UnitTypeId.SUPPLYDEPOT: {AbilityId.MORPH_SUPPLYDEPOT_LOWER},
    UnitTypeId.SUPPLYDEPOTLOWERED: {AbilityId.MORPH_SUPPLYDEPOT_RAISE},
    UnitTypeId.SWARMHOSTBURROWEDMP: {AbilityId.EFFECT_SPAWNLOCUSTS, AbilityId.SMART},
    UnitTypeId.SWARMHOSTMP: {
        AbilityId.BURROWDOWN_SWARMHOST,
        AbilityId.EFFECT_SPAWNLOCUSTS,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.TECHLAB: {
        AbilityId.BARRACKSTECHLABRESEARCH_STIMPACK,
        AbilityId.RESEARCH_BANSHEECLOAKINGFIELD,
        AbilityId.RESEARCH_COMBATSHIELD,
        AbilityId.RESEARCH_CONCUSSIVESHELLS,
        AbilityId.RESEARCH_DRILLINGCLAWS,
        AbilityId.RESEARCH_INFERNALPREIGNITER,
        AbilityId.RESEARCH_RAVENCORVIDREACTOR,
    },
    UnitTypeId.TEMPEST: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.TEMPLARARCHIVE: {AbilityId.RESEARCH_PSISTORM},
    UnitTypeId.THOR: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPH_THORHIGHIMPACTMODE,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.THORAP: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPH_THOREXPLOSIVEMODE,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.TWILIGHTCOUNCIL: {
        AbilityId.RESEARCH_ADEPTRESONATINGGLAIVES,
        AbilityId.RESEARCH_BLINK,
        AbilityId.RESEARCH_CHARGE,
    },
    UnitTypeId.ULTRALISK: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BURROWDOWN_ULTRALISK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.ULTRALISKBURROWED: {AbilityId.BURROWUP_ULTRALISK},
    UnitTypeId.ULTRALISKCAVERN: {AbilityId.RESEARCH_ANABOLICSYNTHESIS, AbilityId.RESEARCH_CHITINOUSPLATING},
    UnitTypeId.VIKINGASSAULT: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPH_VIKINGFIGHTERMODE,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.VIKINGFIGHTER: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPH_VIKINGASSAULTMODE,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.VIPER: {
        AbilityId.BLINDINGCLOUD_BLINDINGCLOUD,
        AbilityId.EFFECT_ABDUCT,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PARASITICBOMB_PARASITICBOMB,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
        AbilityId.VIPERCONSUMESTRUCTURE_VIPERCONSUME,
    },
    UnitTypeId.VOIDMPIMMORTALREVIVECORPSE: {
        AbilityId.RALLY_BUILDING,
        AbilityId.SMART,
        AbilityId.VOIDMPIMMORTALREVIVEREBUILD_IMMORTAL,
    },
    UnitTypeId.VOIDRAY: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.EFFECT_VOIDRAYPRISMATICALIGNMENT,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.WARHOUND: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
        AbilityId.TORNADOMISSILE_TORNADOMISSILE,
    },
    UnitTypeId.WARPGATE: {
        AbilityId.MORPH_GATEWAY,
        AbilityId.SMART,
        AbilityId.TRAINWARP_ADEPT,
        AbilityId.WARPGATETRAIN_DARKTEMPLAR,
        AbilityId.WARPGATETRAIN_HIGHTEMPLAR,
        AbilityId.WARPGATETRAIN_SENTRY,
        AbilityId.WARPGATETRAIN_STALKER,
        AbilityId.WARPGATETRAIN_ZEALOT,
    },
    UnitTypeId.WARPPRISM: {
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.LOAD_WARPPRISM,
        AbilityId.MORPH_WARPPRISMPHASINGMODE,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.WARPPRISMPHASING: {
        AbilityId.LOAD_WARPPRISM,
        AbilityId.MORPH_WARPPRISMTRANSPORTMODE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.WIDOWMINE: {
        AbilityId.BURROWDOWN_WIDOWMINE,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.WIDOWMINEBURROWED: {
        AbilityId.BURROWUP_WIDOWMINE,
        AbilityId.SMART,
        AbilityId.WIDOWMINEATTACK_WIDOWMINEATTACK,
    },
    UnitTypeId.ZEALOT: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.EFFECT_CHARGE,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.ZERGLING: {
        AbilityId.ATTACK_ATTACK,
        AbilityId.BURROWDOWN_ZERGLING,
        AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPHTOBANELING_BANELING,
        AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL,
        AbilityId.SMART,
        AbilityId.STOP_STOP,
    },
    UnitTypeId.ZERGLINGBURROWED: {AbilityId.BURROWUP_ZERGLING},
}



# pyre-ignore-all-errors[6, 11, 16, 29]
from __future__ import annotations

from abc import ABC
from pathlib import Path

from s2clientprotocol import sc2api_pb2
from sc2.bot_ai import BotAI
from sc2.data import AIBuild, Difficulty, PlayerType, Race


class AbstractPlayer(ABC):
    def __init__(
        self,
        p_type: PlayerType,
        race: Race | None = None,
        name: str | None = None,
        difficulty: Difficulty | None = None,
        ai_build: AIBuild | None = None,
        fullscreen: bool = False,
    ) -> None:
        assert isinstance(p_type, PlayerType), f"p_type is of type {type(p_type)}"
        assert name is None or isinstance(name, str), f"name is of type {type(name)}"

        self.name = name
        self.type = p_type
        self.fullscreen = fullscreen
        if race is not None:
            self.race = race
        if p_type == PlayerType.Computer:
            assert isinstance(difficulty, Difficulty), f"difficulty is of type {type(difficulty)}"
            # Workaround, proto information does not carry ai_build info
            # We cant set that in the Player classmethod
            assert ai_build is None or isinstance(ai_build, AIBuild), f"ai_build is of type {type(ai_build)}"
            self.difficulty = difficulty
            self.ai_build = ai_build

        elif p_type == PlayerType.Observer:
            assert race is None
            assert difficulty is None
            assert ai_build is None

        else:
            assert isinstance(race, Race), f"race is of type {type(race)}"
            assert difficulty is None
            assert ai_build is None

    @property
    def needs_sc2(self) -> bool:
        return not isinstance(self, Computer)


class Human(AbstractPlayer):
    def __init__(self, race: Race, name: str | None = None, fullscreen: bool = False) -> None:
        super().__init__(PlayerType.Participant, race, name=name, fullscreen=fullscreen)

    def __str__(self) -> str:
        if self.name is not None:
            return f"Human({self.race._name_}, name={self.name!r})"
        return f"Human({self.race._name_})"


class Bot(AbstractPlayer):
    def __init__(self, race: Race, ai: BotAI, name: str | None = None, fullscreen: bool = False) -> None:
        """
        AI can be None if this player object is just used to inform the
        server about player types.
        """
        assert isinstance(ai, BotAI) or ai is None, f"ai is of type {type(ai)}, inherit BotAI from bot_ai.py"
        super().__init__(PlayerType.Participant, race, name=name, fullscreen=fullscreen)
        self.ai = ai

    def __str__(self) -> str:
        if self.name is not None:
            return f"Bot {self.ai.__class__.__name__}({self.race._name_}), name={self.name!r})"
        return f"Bot {self.ai.__class__.__name__}({self.race._name_})"


class Computer(AbstractPlayer):
    def __init__(
        self, race: Race, difficulty: Difficulty = Difficulty.Easy, ai_build: AIBuild = AIBuild.RandomBuild
    ) -> None:
        super().__init__(PlayerType.Computer, race, difficulty=difficulty, ai_build=ai_build)

    def __str__(self) -> str:
        return f"Computer {self.difficulty._name_}({self.race._name_}, {self.ai_build.name})"


class Observer(AbstractPlayer):
    def __init__(self) -> None:
        super().__init__(PlayerType.Observer)

    def __str__(self) -> str:
        return "Observer"


class Player(AbstractPlayer):
    def __init__(
        self,
        player_id: int,
        p_type: PlayerType,
        requested_race: Race,
        difficulty: Difficulty | None = None,
        actual_race: Race | None = None,
        name: str | None = None,
        ai_build: AIBuild | None = None,
    ) -> None:
        super().__init__(p_type, requested_race, difficulty=difficulty, name=name, ai_build=ai_build)
        self.id: int = player_id
        self.actual_race: Race | None = actual_race

    @classmethod
    def from_proto(cls, proto: sc2api_pb2.PlayerInfo) -> Player:
        if PlayerType(proto.type) == PlayerType.Observer:
            return cls(proto.player_id, PlayerType(proto.type), None, None, None)
        return cls(
            proto.player_id,
            PlayerType(proto.type),
            Race(proto.race_requested),
            Difficulty(proto.difficulty) if proto.HasField("difficulty") else None,
            Race(proto.race_actual) if proto.HasField("race_actual") else None,
            proto.player_name if proto.HasField("player_name") else None,
        )


class BotProcess(AbstractPlayer):
    """
    Class for handling bots launched externally, including non-python bots.
    Default parameters comply with sc2ai and aiarena ladders.

    :param path: the executable file's path
    :param launch_list: list of strings that launches the bot e.g. ["python", "run.py"] or ["run.exe"]
    :param race: bot's race
    :param name: bot's name
    :param sc2port_arg: the accepted argument name for the port of the sc2 instance to listen to
    :param hostaddress_arg: the accepted argument name for the address of the sc2 instance to listen to
    :param match_arg: the accepted argument name for the starting port to generate a portconfig from
    :param realtime_arg: the accepted argument name for specifying realtime
    :param other_args: anything else that is needed

    e.g. to call a bot capable of running on the bot ladders:
        BotProcess(os.getcwd(), "python run.py", Race.Terran, "INnoVation")
    """

    def __init__(
        self,
        path: str | Path,
        launch_list: list[str],
        race: Race,
        name: str | None = None,
        sc2port_arg: str = "--GamePort",
        hostaddress_arg: str = "--LadderServer",
        match_arg: str = "--StartPort",
        realtime_arg: str = "--RealTime",
        other_args: str | None = None,
        stdout: str | None = None,
    ) -> None:
        super().__init__(PlayerType.Participant, race, name=name)
        assert Path(path).exists()
        self.path = path
        self.launch_list = launch_list
        self.sc2port_arg = sc2port_arg
        self.match_arg = match_arg
        self.hostaddress_arg = hostaddress_arg
        self.realtime_arg = realtime_arg
        self.other_args = other_args
        self.stdout = stdout

    def __repr__(self) -> str:
        if self.name is not None:
            return f"Bot {self.name}({self.race.name} from {self.launch_list})"
        return f"Bot({self.race.name} from {self.launch_list})"

    def cmd_line(
        self, sc2port: int | str, matchport: int | str | None, hostaddress: str, realtime: bool = False
    ) -> list[str]:
        """

        :param sc2port: the port that the launched sc2 instance listens to
        :param matchport: some starting port that both bots use to generate identical portconfigs.
                Note: This will not be sent if playing vs computer
        :param hostaddress: the address the sc2 instances used
        :param realtime: 1 or 0, indicating whether the match is played in realtime or not
        :return: string that will be used to start the bot's process
        """
        cmd_line = [
            *self.launch_list,
            self.sc2port_arg,
            str(sc2port),
            self.hostaddress_arg,
            hostaddress,
        ]
        if matchport is not None:
            cmd_line.extend([self.match_arg, str(matchport)])
        if self.other_args is not None:
            cmd_line.append(self.other_args)
        if realtime:
            cmd_line.extend([self.realtime_arg])
        return cmd_line



# THIS FILE WAS AUTOMATICALLY GENERATED BY "generate_dicts_from_data_json.py" DO NOT CHANGE MANUALLY!
# ANY CHANGE WILL BE OVERWRITTEN

# from sc2.ids.buff_id import BuffId
# from sc2.ids.effect_id import EffectId
from typing import Union

from sc2.ids.ability_id import AbilityId
from sc2.ids.unit_typeid import UnitTypeId
from sc2.ids.upgrade_id import UpgradeId

RESEARCH_INFO: dict[UnitTypeId, dict[UpgradeId, dict[str, Union[AbilityId, bool, UnitTypeId, UpgradeId]]]] = {
    UnitTypeId.ARMORY: {
        UpgradeId.TERRANSHIPWEAPONSLEVEL1: {"ability": AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL1},
        UpgradeId.TERRANSHIPWEAPONSLEVEL2: {
            "ability": AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL2,
            "required_upgrade": UpgradeId.TERRANSHIPWEAPONSLEVEL1,
        },
        UpgradeId.TERRANSHIPWEAPONSLEVEL3: {
            "ability": AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL3,
            "required_upgrade": UpgradeId.TERRANSHIPWEAPONSLEVEL2,
        },
        UpgradeId.TERRANVEHICLEANDSHIPARMORSLEVEL1: {
            "ability": AbilityId.ARMORYRESEARCH_TERRANVEHICLEANDSHIPPLATINGLEVEL1
        },
        UpgradeId.TERRANVEHICLEANDSHIPARMORSLEVEL2: {
            "ability": AbilityId.ARMORYRESEARCH_TERRANVEHICLEANDSHIPPLATINGLEVEL2,
            "required_upgrade": UpgradeId.TERRANVEHICLEANDSHIPARMORSLEVEL1,
        },
        UpgradeId.TERRANVEHICLEANDSHIPARMORSLEVEL3: {
            "ability": AbilityId.ARMORYRESEARCH_TERRANVEHICLEANDSHIPPLATINGLEVEL3,
            "required_upgrade": UpgradeId.TERRANVEHICLEANDSHIPARMORSLEVEL2,
        },
        UpgradeId.TERRANVEHICLEWEAPONSLEVEL1: {"ability": AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL1},
        UpgradeId.TERRANVEHICLEWEAPONSLEVEL2: {
            "ability": AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL2,
            "required_upgrade": UpgradeId.TERRANVEHICLEWEAPONSLEVEL1,
        },
        UpgradeId.TERRANVEHICLEWEAPONSLEVEL3: {
            "ability": AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL3,
            "required_upgrade": UpgradeId.TERRANVEHICLEWEAPONSLEVEL2,
        },
    },
    UnitTypeId.BANELINGNEST: {
        UpgradeId.CENTRIFICALHOOKS: {
            "ability": AbilityId.RESEARCH_CENTRIFUGALHOOKS,
            "required_building": UnitTypeId.LAIR,
        }
    },
    UnitTypeId.BARRACKSTECHLAB: {
        UpgradeId.PUNISHERGRENADES: {"ability": AbilityId.RESEARCH_CONCUSSIVESHELLS},
        UpgradeId.SHIELDWALL: {"ability": AbilityId.RESEARCH_COMBATSHIELD},
        UpgradeId.STIMPACK: {"ability": AbilityId.BARRACKSTECHLABRESEARCH_STIMPACK},
    },
    UnitTypeId.CYBERNETICSCORE: {
        UpgradeId.PROTOSSAIRARMORSLEVEL1: {
            "ability": AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRARMORLEVEL1,
            "requires_power": True,
        },
        UpgradeId.PROTOSSAIRARMORSLEVEL2: {
            "ability": AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRARMORLEVEL2,
            "required_building": UnitTypeId.FLEETBEACON,
            "required_upgrade": UpgradeId.PROTOSSAIRARMORSLEVEL1,
            "requires_power": True,
        },
        UpgradeId.PROTOSSAIRARMORSLEVEL3: {
            "ability": AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRARMORLEVEL3,
            "required_building": UnitTypeId.FLEETBEACON,
            "required_upgrade": UpgradeId.PROTOSSAIRARMORSLEVEL2,
            "requires_power": True,
        },
        UpgradeId.PROTOSSAIRWEAPONSLEVEL1: {
            "ability": AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL1,
            "requires_power": True,
        },
        UpgradeId.PROTOSSAIRWEAPONSLEVEL2: {
            "ability": AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL2,
            "required_building": UnitTypeId.FLEETBEACON,
            "required_upgrade": UpgradeId.PROTOSSAIRWEAPONSLEVEL1,
            "requires_power": True,
        },
        UpgradeId.PROTOSSAIRWEAPONSLEVEL3: {
            "ability": AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL3,
            "required_building": UnitTypeId.FLEETBEACON,
            "required_upgrade": UpgradeId.PROTOSSAIRWEAPONSLEVEL2,
            "requires_power": True,
        },
        UpgradeId.WARPGATERESEARCH: {"ability": AbilityId.RESEARCH_WARPGATE, "requires_power": True},
    },
    UnitTypeId.DARKSHRINE: {
        UpgradeId.DARKTEMPLARBLINKUPGRADE: {"ability": AbilityId.RESEARCH_SHADOWSTRIKE, "requires_power": True}
    },
    UnitTypeId.ENGINEERINGBAY: {
        UpgradeId.HISECAUTOTRACKING: {"ability": AbilityId.RESEARCH_HISECAUTOTRACKING},
        UpgradeId.TERRANBUILDINGARMOR: {"ability": AbilityId.RESEARCH_TERRANSTRUCTUREARMORUPGRADE},
        UpgradeId.TERRANINFANTRYARMORSLEVEL1: {"ability": AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL1},
        UpgradeId.TERRANINFANTRYARMORSLEVEL2: {
            "ability": AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL2,
            "required_building": UnitTypeId.ARMORY,
            "required_upgrade": UpgradeId.TERRANINFANTRYARMORSLEVEL1,
        },
        UpgradeId.TERRANINFANTRYARMORSLEVEL3: {
            "ability": AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL3,
            "required_building": UnitTypeId.ARMORY,
            "required_upgrade": UpgradeId.TERRANINFANTRYARMORSLEVEL2,
        },
        UpgradeId.TERRANINFANTRYWEAPONSLEVEL1: {
            "ability": AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL1
        },
        UpgradeId.TERRANINFANTRYWEAPONSLEVEL2: {
            "ability": AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL2,
            "required_building": UnitTypeId.ARMORY,
            "required_upgrade": UpgradeId.TERRANINFANTRYWEAPONSLEVEL1,
        },
        UpgradeId.TERRANINFANTRYWEAPONSLEVEL3: {
            "ability": AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL3,
            "required_building": UnitTypeId.ARMORY,
            "required_upgrade": UpgradeId.TERRANINFANTRYWEAPONSLEVEL2,
        },
    },
    UnitTypeId.EVOLUTIONCHAMBER: {
        UpgradeId.ZERGGROUNDARMORSLEVEL1: {"ability": AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL1},
        UpgradeId.ZERGGROUNDARMORSLEVEL2: {
            "ability": AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL2,
            "required_building": UnitTypeId.LAIR,
            "required_upgrade": UpgradeId.ZERGGROUNDARMORSLEVEL1,
        },
        UpgradeId.ZERGGROUNDARMORSLEVEL3: {
            "ability": AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL3,
            "required_building": UnitTypeId.HIVE,
            "required_upgrade": UpgradeId.ZERGGROUNDARMORSLEVEL2,
        },
        UpgradeId.ZERGMELEEWEAPONSLEVEL1: {"ability": AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL1},
        UpgradeId.ZERGMELEEWEAPONSLEVEL2: {
            "ability": AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL2,
            "required_building": UnitTypeId.LAIR,
            "required_upgrade": UpgradeId.ZERGMELEEWEAPONSLEVEL1,
        },
        UpgradeId.ZERGMELEEWEAPONSLEVEL3: {
            "ability": AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL3,
            "required_building": UnitTypeId.HIVE,
            "required_upgrade": UpgradeId.ZERGMELEEWEAPONSLEVEL2,
        },
        UpgradeId.ZERGMISSILEWEAPONSLEVEL1: {"ability": AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL1},
        UpgradeId.ZERGMISSILEWEAPONSLEVEL2: {
            "ability": AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL2,
            "required_building": UnitTypeId.LAIR,
            "required_upgrade": UpgradeId.ZERGMISSILEWEAPONSLEVEL1,
        },
        UpgradeId.ZERGMISSILEWEAPONSLEVEL3: {
            "ability": AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL3,
            "required_building": UnitTypeId.HIVE,
            "required_upgrade": UpgradeId.ZERGMISSILEWEAPONSLEVEL2,
        },
    },
    UnitTypeId.FACTORYTECHLAB: {
        UpgradeId.CYCLONELOCKONDAMAGEUPGRADE: {"ability": AbilityId.RESEARCH_CYCLONELOCKONDAMAGE},
        UpgradeId.DRILLCLAWS: {"ability": AbilityId.RESEARCH_DRILLINGCLAWS, "required_building": UnitTypeId.ARMORY},
        UpgradeId.HIGHCAPACITYBARRELS: {"ability": AbilityId.RESEARCH_INFERNALPREIGNITER},
        UpgradeId.SMARTSERVOS: {"ability": AbilityId.RESEARCH_SMARTSERVOS, "required_building": UnitTypeId.ARMORY},
    },
    UnitTypeId.FLEETBEACON: {
        UpgradeId.PHOENIXRANGEUPGRADE: {
            "ability": AbilityId.RESEARCH_PHOENIXANIONPULSECRYSTALS,
            "requires_power": True,
        },
        UpgradeId.TEMPESTGROUNDATTACKUPGRADE: {
            "ability": AbilityId.FLEETBEACONRESEARCH_TEMPESTRESEARCHGROUNDATTACKUPGRADE,
            "requires_power": True,
        },
        UpgradeId.VOIDRAYSPEEDUPGRADE: {
            "ability": AbilityId.FLEETBEACONRESEARCH_RESEARCHVOIDRAYSPEEDUPGRADE,
            "requires_power": True,
        },
    },
    UnitTypeId.FORGE: {
        UpgradeId.PROTOSSGROUNDARMORSLEVEL1: {
            "ability": AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL1,
            "requires_power": True,
        },
        UpgradeId.PROTOSSGROUNDARMORSLEVEL2: {
            "ability": AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL2,
            "required_building": UnitTypeId.TWILIGHTCOUNCIL,
            "required_upgrade": UpgradeId.PROTOSSGROUNDARMORSLEVEL1,
            "requires_power": True,
        },
        UpgradeId.PROTOSSGROUNDARMORSLEVEL3: {
            "ability": AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL3,
            "required_building": UnitTypeId.TWILIGHTCOUNCIL,
            "required_upgrade": UpgradeId.PROTOSSGROUNDARMORSLEVEL2,
            "requires_power": True,
        },
        UpgradeId.PROTOSSGROUNDWEAPONSLEVEL1: {
            "ability": AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL1,
            "requires_power": True,
        },
        UpgradeId.PROTOSSGROUNDWEAPONSLEVEL2: {
            "ability": AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL2,
            "required_building": UnitTypeId.TWILIGHTCOUNCIL,
            "required_upgrade": UpgradeId.PROTOSSGROUNDWEAPONSLEVEL1,
            "requires_power": True,
        },
        UpgradeId.PROTOSSGROUNDWEAPONSLEVEL3: {
            "ability": AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL3,
            "required_building": UnitTypeId.TWILIGHTCOUNCIL,
            "required_upgrade": UpgradeId.PROTOSSGROUNDWEAPONSLEVEL2,
            "requires_power": True,
        },
        UpgradeId.PROTOSSSHIELDSLEVEL1: {
            "ability": AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL1,
            "requires_power": True,
        },
        UpgradeId.PROTOSSSHIELDSLEVEL2: {
            "ability": AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL2,
            "required_building": UnitTypeId.TWILIGHTCOUNCIL,
            "required_upgrade": UpgradeId.PROTOSSSHIELDSLEVEL1,
            "requires_power": True,
        },
        UpgradeId.PROTOSSSHIELDSLEVEL3: {
            "ability": AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL3,
            "required_building": UnitTypeId.TWILIGHTCOUNCIL,
            "required_upgrade": UpgradeId.PROTOSSSHIELDSLEVEL2,
            "requires_power": True,
        },
    },
    UnitTypeId.FUSIONCORE: {
        UpgradeId.BATTLECRUISERENABLESPECIALIZATIONS: {"ability": AbilityId.RESEARCH_BATTLECRUISERWEAPONREFIT},
        UpgradeId.LIBERATORAGRANGEUPGRADE: {"ability": AbilityId.FUSIONCORERESEARCH_RESEARCHBALLISTICRANGE},
        UpgradeId.MEDIVACCADUCEUSREACTOR: {"ability": AbilityId.FUSIONCORERESEARCH_RESEARCHMEDIVACENERGYUPGRADE},
    },
    UnitTypeId.GHOSTACADEMY: {UpgradeId.PERSONALCLOAKING: {"ability": AbilityId.RESEARCH_PERSONALCLOAKING}},
    UnitTypeId.GREATERSPIRE: {
        UpgradeId.ZERGFLYERARMORSLEVEL1: {"ability": AbilityId.RESEARCH_ZERGFLYERARMORLEVEL1},
        UpgradeId.ZERGFLYERARMORSLEVEL2: {
            "ability": AbilityId.RESEARCH_ZERGFLYERARMORLEVEL2,
            "required_building": UnitTypeId.LAIR,
            "required_upgrade": UpgradeId.ZERGFLYERARMORSLEVEL1,
        },
        UpgradeId.ZERGFLYERARMORSLEVEL3: {
            "ability": AbilityId.RESEARCH_ZERGFLYERARMORLEVEL3,
            "required_building": UnitTypeId.HIVE,
            "required_upgrade": UpgradeId.ZERGFLYERARMORSLEVEL2,
        },
        UpgradeId.ZERGFLYERWEAPONSLEVEL1: {"ability": AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL1},
        UpgradeId.ZERGFLYERWEAPONSLEVEL2: {
            "ability": AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL2,
            "required_building": UnitTypeId.LAIR,
            "required_upgrade": UpgradeId.ZERGFLYERWEAPONSLEVEL1,
        },
        UpgradeId.ZERGFLYERWEAPONSLEVEL3: {
            "ability": AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL3,
            "required_building": UnitTypeId.HIVE,
            "required_upgrade": UpgradeId.ZERGFLYERWEAPONSLEVEL2,
        },
    },
    UnitTypeId.HATCHERY: {
        UpgradeId.BURROW: {"ability": AbilityId.RESEARCH_BURROW},
        UpgradeId.OVERLORDSPEED: {"ability": AbilityId.RESEARCH_PNEUMATIZEDCARAPACE},
    },
    UnitTypeId.HIVE: {
        UpgradeId.BURROW: {"ability": AbilityId.RESEARCH_BURROW},
        UpgradeId.OVERLORDSPEED: {"ability": AbilityId.RESEARCH_PNEUMATIZEDCARAPACE},
    },
    UnitTypeId.HYDRALISKDEN: {
        UpgradeId.EVOLVEGROOVEDSPINES: {"ability": AbilityId.RESEARCH_GROOVEDSPINES},
        UpgradeId.EVOLVEMUSCULARAUGMENTS: {"ability": AbilityId.RESEARCH_MUSCULARAUGMENTS},
        UpgradeId.FRENZY: {
            "ability": AbilityId.HYDRALISKDENRESEARCH_RESEARCHFRENZY,
            "required_building": UnitTypeId.HIVE,
        },
    },
    UnitTypeId.INFESTATIONPIT: {UpgradeId.NEURALPARASITE: {"ability": AbilityId.RESEARCH_NEURALPARASITE}},
    UnitTypeId.LAIR: {
        UpgradeId.BURROW: {"ability": AbilityId.RESEARCH_BURROW},
        UpgradeId.OVERLORDSPEED: {"ability": AbilityId.RESEARCH_PNEUMATIZEDCARAPACE},
    },
    UnitTypeId.LURKERDENMP: {
        UpgradeId.DIGGINGCLAWS: {"ability": AbilityId.RESEARCH_ADAPTIVETALONS, "required_building": UnitTypeId.HIVE},
        UpgradeId.LURKERRANGE: {
            "ability": AbilityId.LURKERDENRESEARCH_RESEARCHLURKERRANGE,
            "required_building": UnitTypeId.HIVE,
        },
    },
    UnitTypeId.ROACHWARREN: {
        UpgradeId.GLIALRECONSTITUTION: {
            "ability": AbilityId.RESEARCH_GLIALREGENERATION,
            "required_building": UnitTypeId.LAIR,
        },
        UpgradeId.TUNNELINGCLAWS: {"ability": AbilityId.RESEARCH_TUNNELINGCLAWS, "required_building": UnitTypeId.LAIR},
    },
    UnitTypeId.ROBOTICSBAY: {
        UpgradeId.EXTENDEDTHERMALLANCE: {"ability": AbilityId.RESEARCH_EXTENDEDTHERMALLANCE, "requires_power": True},
        UpgradeId.GRAVITICDRIVE: {"ability": AbilityId.RESEARCH_GRAVITICDRIVE, "requires_power": True},
        UpgradeId.OBSERVERGRAVITICBOOSTER: {"ability": AbilityId.RESEARCH_GRAVITICBOOSTER, "requires_power": True},
    },
    UnitTypeId.SPAWNINGPOOL: {
        UpgradeId.ZERGLINGATTACKSPEED: {
            "ability": AbilityId.RESEARCH_ZERGLINGADRENALGLANDS,
            "required_building": UnitTypeId.HIVE,
        },
        UpgradeId.ZERGLINGMOVEMENTSPEED: {"ability": AbilityId.RESEARCH_ZERGLINGMETABOLICBOOST},
    },
    UnitTypeId.SPIRE: {
        UpgradeId.ZERGFLYERARMORSLEVEL1: {"ability": AbilityId.RESEARCH_ZERGFLYERARMORLEVEL1},
        UpgradeId.ZERGFLYERARMORSLEVEL2: {
            "ability": AbilityId.RESEARCH_ZERGFLYERARMORLEVEL2,
            "required_building": UnitTypeId.LAIR,
            "required_upgrade": UpgradeId.ZERGFLYERARMORSLEVEL1,
        },
        UpgradeId.ZERGFLYERARMORSLEVEL3: {
            "ability": AbilityId.RESEARCH_ZERGFLYERARMORLEVEL3,
            "required_building": UnitTypeId.HIVE,
            "required_upgrade": UpgradeId.ZERGFLYERARMORSLEVEL2,
        },
        UpgradeId.ZERGFLYERWEAPONSLEVEL1: {"ability": AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL1},
        UpgradeId.ZERGFLYERWEAPONSLEVEL2: {
            "ability": AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL2,
            "required_building": UnitTypeId.LAIR,
            "required_upgrade": UpgradeId.ZERGFLYERWEAPONSLEVEL1,
        },
        UpgradeId.ZERGFLYERWEAPONSLEVEL3: {
            "ability": AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL3,
            "required_building": UnitTypeId.HIVE,
            "required_upgrade": UpgradeId.ZERGFLYERWEAPONSLEVEL2,
        },
    },
    UnitTypeId.STARPORTTECHLAB: {
        UpgradeId.BANSHEECLOAK: {"ability": AbilityId.RESEARCH_BANSHEECLOAKINGFIELD},
        UpgradeId.BANSHEESPEED: {"ability": AbilityId.RESEARCH_BANSHEEHYPERFLIGHTROTORS},
        UpgradeId.INTERFERENCEMATRIX: {"ability": AbilityId.STARPORTTECHLABRESEARCH_RESEARCHRAVENINTERFERENCEMATRIX},
    },
    UnitTypeId.TEMPLARARCHIVE: {
        UpgradeId.PSISTORMTECH: {"ability": AbilityId.RESEARCH_PSISTORM, "requires_power": True}
    },
    UnitTypeId.TWILIGHTCOUNCIL: {
        UpgradeId.ADEPTPIERCINGATTACK: {"ability": AbilityId.RESEARCH_ADEPTRESONATINGGLAIVES, "requires_power": True},
        UpgradeId.BLINKTECH: {"ability": AbilityId.RESEARCH_BLINK, "requires_power": True},
        UpgradeId.CHARGE: {"ability": AbilityId.RESEARCH_CHARGE, "requires_power": True},
    },
    UnitTypeId.ULTRALISKCAVERN: {
        UpgradeId.ANABOLICSYNTHESIS: {"ability": AbilityId.RESEARCH_ANABOLICSYNTHESIS},
        UpgradeId.CHITINOUSPLATING: {"ability": AbilityId.RESEARCH_CHITINOUSPLATING},
    },
}



from __future__ import annotations

from enum import Enum

from google.protobuf.message import Message

class ScoreType(Enum):
    Curriculum: int
    Melee: int

class Score(Message):
    score_type: int
    score: int
    score_details: ScoreDetails
    def __init__(
        self,
        score_type: int = ...,
        score: int = ...,
        score_details: ScoreDetails = ...,
    ) -> None: ...

class CategoryScoreDetails(Message):
    none: float
    army: float
    economy: float
    technology: float
    upgrade: float
    def __init__(
        self,
        none: float = ...,
        army: float = ...,
        economy: float = ...,
        technology: float = ...,
        upgrade: float = ...,
    ) -> None: ...

class VitalScoreDetails(Message):
    life: float
    shields: float
    energy: float
    def __init__(
        self,
        life: float = ...,
        shields: float = ...,
        energy: float = ...,
    ) -> None: ...

class ScoreDetails(Message):
    idle_production_time: float
    idle_worker_time: float
    total_value_units: float
    total_value_structures: float
    killed_value_units: float
    killed_value_structures: float
    collected_minerals: float
    collected_vespene: float
    collection_rate_minerals: float
    collection_rate_vespene: float
    spent_minerals: float
    spent_vespene: float
    food_used: CategoryScoreDetails
    killed_minerals: CategoryScoreDetails
    killed_vespene: CategoryScoreDetails
    lost_minerals: CategoryScoreDetails
    lost_vespene: CategoryScoreDetails
    friendly_fire_minerals: CategoryScoreDetails
    friendly_fire_vespene: CategoryScoreDetails
    used_minerals: CategoryScoreDetails
    used_vespene: CategoryScoreDetails
    total_used_minerals: CategoryScoreDetails
    total_used_vespene: CategoryScoreDetails
    total_damage_dealt: VitalScoreDetails
    total_damage_taken: VitalScoreDetails
    total_healed: VitalScoreDetails
    current_apm: float
    current_effective_apm: float
    def __init__(
        self,
        idle_production_time: float = ...,
        idle_worker_time: float = ...,
        total_value_units: float = ...,
        total_value_structures: float = ...,
        killed_value_units: float = ...,
        killed_value_structures: float = ...,
        collected_minerals: float = ...,
        collected_vespene: float = ...,
        collection_rate_minerals: float = ...,
        collection_rate_vespene: float = ...,
        spent_minerals: float = ...,
        spent_vespene: float = ...,
        food_used: CategoryScoreDetails = ...,
        killed_minerals: CategoryScoreDetails = ...,
        killed_vespene: CategoryScoreDetails = ...,
        lost_minerals: CategoryScoreDetails = ...,
        lost_vespene: CategoryScoreDetails = ...,
        friendly_fire_minerals: CategoryScoreDetails = ...,
        friendly_fire_vespene: CategoryScoreDetails = ...,
        used_minerals: CategoryScoreDetails = ...,
        used_vespene: CategoryScoreDetails = ...,
        total_used_minerals: CategoryScoreDetails = ...,
        total_used_vespene: CategoryScoreDetails = ...,
        total_damage_dealt: VitalScoreDetails = ...,
        total_damage_taken: VitalScoreDetails = ...,
        total_healed: VitalScoreDetails = ...,
        current_apm: float = ...,
        current_effective_apm: float = ...,
    ) -> None: ...



# https://github.com/Blizzard/s2client-proto/blob/bff45dae1fc685e6acbaae084670afb7d1c0832c/s2clientprotocol/common.proto
from enum import Enum

from google.protobuf.message import Message

class AvailableAbility(Message):
    ability_id: int
    requires_point: bool
    def __init__(self, ability_id: int = ..., requires_point: bool = ...) -> None: ...

class ImageData(Message):
    bits_per_pixel: int
    size: Size2DI
    data: bytes
    def __init__(self, bits_per_pixel: int = ..., size: Size2DI = ..., data: bytes = ...) -> None: ...

class PointI(Message):
    x: int
    y: int
    def __init__(self, x: int = ..., y: int = ...) -> None: ...

class RectangleI(Message):
    p0: PointI
    p1: PointI
    def __init__(self, p0: PointI = ..., p1: PointI = ...) -> None: ...

class Point2D(Message):
    x: float
    y: float
    def __init__(self, x: float = ..., y: float = ...) -> None: ...

class Point(Message):
    x: float
    y: float
    z: float
    def __init__(self, x: float = ..., y: float = ..., z: float = ...) -> None: ...

class Size2DI(Message):
    x: int
    y: int
    def __init__(self, x: int = ..., y: int = ...) -> None: ...

class Race(Enum):
    NoRace: int
    Terran: int
    Zerg: int
    Protoss: int
    Random: int



# pyre-ignore-all-errors[16, 29]
from __future__ import annotations

import asyncio
import os
import platform
import subprocess
import time
import traceback
from pathlib import Path

from aiohttp import WSMsgType, web
from aiohttp.web_ws import WebSocketResponse
from loguru import logger

# pyre-fixme[21]
from s2clientprotocol import sc2api_pb2 as sc_pb
from sc2.controller import Controller
from sc2.data import Result, Status
from sc2.player import BotProcess


class Proxy:
    """
    Class for handling communication between sc2 and an external bot.
    This "middleman" is needed for enforcing time limits, collecting results, and closing things properly.
    """

    def __init__(
        self,
        controller: Controller,
        player: BotProcess,
        proxyport: int,
        game_time_limit: int | None = None,
        realtime: bool = False,
    ) -> None:
        self.controller = controller
        self.player = player
        self.port = proxyport
        self.timeout_loop = game_time_limit * 22.4 if game_time_limit else None
        self.realtime = realtime
        logger.debug(
            f"Proxy Inited with ctrl {controller}({controller._process._port}), player {player}, proxyport {proxyport}, lim {game_time_limit}"
        )

        self.result = None
        self.player_id: int | None = None
        self.done = False

    async def parse_request(self, msg) -> None:
        request = sc_pb.Request()
        request.ParseFromString(msg.data)
        if request.HasField("quit"):
            request = sc_pb.Request(leave_game=sc_pb.RequestLeaveGame())
        if request.HasField("leave_game"):
            if self.controller._status == Status.in_game:
                logger.info(f"Proxy: player {self.player.name}({self.player_id}) surrenders")
                self.result = {self.player_id: Result.Defeat}
            elif self.controller._status == Status.ended:
                await self.get_response()
        elif request.HasField("join_game") and not request.join_game.HasField("player_name"):
            request.join_game.player_name = self.player.name
        await self.controller._ws.send_bytes(request.SerializeToString())

    # TODO Catching too general exception Exception (broad-except)

    async def get_response(self):
        response_bytes = None
        try:
            response_bytes = await self.controller._ws.receive_bytes()
        except TypeError as e:
            logger.exception("Cannot receive: SC2 Connection already closed.")
            tb = traceback.format_exc()
            logger.error(f"Exception {e}: {tb}")
        except asyncio.CancelledError:
            logger.info(f"Proxy({self.player.name}), caught receive from sc2")
            try:
                x = await self.controller._ws.receive_bytes()
                if response_bytes is None:
                    response_bytes = x
            except (asyncio.CancelledError, asyncio.TimeoutError, Exception) as e:
                logger.exception(f"Exception {e}")
        except Exception as e:
            logger.exception(f"Caught unknown exception: {e}")
        return response_bytes

    async def parse_response(self, response_bytes):
        response = sc_pb.Response()
        response.ParseFromString(response_bytes)

        if not response.HasField("status"):
            logger.critical("Proxy: RESPONSE HAS NO STATUS {response}")
        else:
            new_status = Status(response.status)
            if new_status != self.controller._status:
                logger.info(f"Controller({self.player.name}): {self.controller._status}->{new_status}")
                self.controller._status = new_status

        if self.player_id is None and response.HasField("join_game"):
            self.player_id = response.join_game.player_id
            logger.info(f"Proxy({self.player.name}): got join_game for {self.player_id}")

        if self.result is None and response.HasField("observation"):
            obs: sc_pb.ResponseObservation = response.observation
            if obs.player_result:
                self.result = {pr.player_id: Result(pr.result) for pr in obs.player_result}
            elif self.timeout_loop and obs.HasField("observation") and obs.observation.game_loop > self.timeout_loop:
                self.result = {i: Result.Tie for i in range(1, 3)}  # noqa: C420
                logger.info(f"Proxy({self.player.name}) timing out")
                act = [sc_pb.Action(action_chat=sc_pb.ActionChat(message="Proxy: Timing out"))]
                await self.controller._execute(action=sc_pb.RequestAction(actions=act))
        return response

    async def get_result(self) -> None:
        try:
            res = await self.controller.ping()
            if res.status in {Status.in_game, Status.in_replay, Status.ended}:
                res = await self.controller._execute(observation=sc_pb.RequestObservation())
                if res.HasField("observation") and res.observation.player_result:
                    self.result = {pr.player_id: Result(pr.result) for pr in res.observation.player_result}

        # TODO Catching too general exception Exception (broad-except)
        except Exception as e:
            logger.exception(f"Caught unknown exception: {e}")

    async def proxy_handler(self, request) -> WebSocketResponse:
        bot_ws = web.WebSocketResponse(receive_timeout=30)
        await bot_ws.prepare(request)
        try:
            async for msg in bot_ws:
                if msg.data is None:
                    raise TypeError(f"data is None, {msg}")
                if msg.data and msg.type == WSMsgType.BINARY:
                    await self.parse_request(msg)

                    response_bytes = await self.get_response()
                    if response_bytes is None:
                        raise ConnectionError("Could not get response_bytes")

                    new_response = await self.parse_response(response_bytes)
                    await bot_ws.send_bytes(new_response.SerializeToString())

                elif msg.type == WSMsgType.CLOSED:
                    logger.error("Client shutdown")
                else:
                    logger.error("Incorrect message type")

        # TODO Catching too general exception Exception (broad-except)
        except Exception as e:
            logger.exception(f"Caught unknown exception: {e}")
            ignored_errors = {ConnectionError, asyncio.CancelledError}
            if not any(isinstance(e, E) for E in ignored_errors):
                tb = traceback.format_exc()
                logger.info(f"Proxy({self.player.name}): Caught {e} traceback: {tb}")
        finally:
            try:
                if self.controller._status in {Status.in_game, Status.in_replay}:
                    await self.controller._execute(leave_game=sc_pb.RequestLeaveGame())
                await bot_ws.close()

            # TODO Catching too general exception Exception (broad-except)
            except Exception as e:
                logger.exception(f"Caught unknown exception during surrender: {e}")
            self.done = True
        return bot_ws

    async def play_with_proxy(self, startport):
        logger.info(f"Proxy({self.port}): Starting app")
        app = web.Application()
        app.router.add_route("GET", "/sc2api", self.proxy_handler)
        apprunner = web.AppRunner(app, access_log=None)
        await apprunner.setup()
        appsite = web.TCPSite(apprunner, self.controller._process._host, self.port)
        await appsite.start()

        subproc_args = {"cwd": str(self.player.path), "stderr": subprocess.STDOUT}
        if platform.system() == "Linux":
            subproc_args["preexec_fn"] = os.setpgrp
        elif platform.system() == "Windows":
            subproc_args["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP

        player_command_line = self.player.cmd_line(self.port, startport, self.controller._process._host, self.realtime)
        logger.info(f"Starting bot with command: {' '.join(player_command_line)}")
        if self.player.stdout is None:
            bot_process = subprocess.Popen(player_command_line, stdout=subprocess.DEVNULL, **subproc_args)
        else:
            with Path(self.player.stdout).open("w+") as out:
                bot_process = subprocess.Popen(player_command_line, stdout=out, **subproc_args)

        while self.result is None:
            bot_alive = bot_process and bot_process.poll() is None
            sc2_alive = self.controller.running
            if self.done or not (bot_alive and sc2_alive):
                logger.info(
                    f"Proxy({self.port}): {self.player.name} died, "
                    f"bot{(not bot_alive) * ' not'} alive, sc2{(not sc2_alive) * ' not'} alive"
                )
                # Maybe its still possible to retrieve a result
                if sc2_alive and not self.done:
                    await self.get_response()
                logger.info(f"Proxy({self.port}): breaking, result {self.result}")
                break
            await asyncio.sleep(5)

        # cleanup
        logger.info(f"({self.port}): cleaning up {self.player!r}")
        for _i in range(3):
            if isinstance(bot_process, subprocess.Popen):
                if bot_process.stdout and not bot_process.stdout.closed:  # should not run anymore
                    logger.info(f"==================output for player {self.player.name}")
                    for line in bot_process.stdout.readlines():
                        logger.opt(raw=True).info(line.decode("utf-8"))
                    bot_process.stdout.close()
                    logger.info("==================")
                bot_process.terminate()
                bot_process.wait()
            time.sleep(0.5)
            if not bot_process or bot_process.poll() is not None:
                break
        else:
            bot_process.terminate()
            bot_process.wait()
        try:
            await apprunner.cleanup()

        # TODO Catching too general exception Exception (broad-except)
        except Exception as e:
            logger.exception(f"Caught unknown exception during cleaning: {e}")
        if isinstance(self.result, dict):
            self.result[None] = None
            return self.result[self.player_id]
        return self.result



# pyre-ignore-all-errors[14, 15, 16]
from __future__ import annotations

import random
from collections.abc import Callable, Generator, Iterable
from itertools import chain
from typing import TYPE_CHECKING, Any

from s2clientprotocol import raw_pb2
from sc2.ids.unit_typeid import UnitTypeId
from sc2.position import Point2
from sc2.unit import Unit

if TYPE_CHECKING:
    from sc2.bot_ai import BotAI


class Units(list[Unit]):
    """A collection of Unit objects. Makes it easy to select units by selectors."""

    @classmethod
    def from_proto(cls, units: list[raw_pb2.Unit], bot_object: BotAI) -> Units:
        return cls((Unit(raw_unit, bot_object=bot_object) for raw_unit in units), bot_object)

    def __init__(self, units: Iterable[Unit], bot_object: BotAI) -> None:
        """
        :param units:
        :param bot_object:
        """
        super().__init__(units)
        self._bot_object = bot_object

    def __call__(self, unit_types: UnitTypeId | Iterable[UnitTypeId]) -> Units:
        """Creates a new mutable Units object from Units or list object.

        :param unit_types:
        """
        return self.of_type(unit_types)

    def __iter__(self) -> Generator[Unit, None, None]:
        return (item for item in super().__iter__())

    def copy(self) -> Units:
        """Creates a new mutable Units object from Units or list object.

        :param units:
        """
        return Units(self, self._bot_object)

    def __or__(self, other: Units) -> Units:
        """
        :param other:
        """
        return Units(
            chain(
                iter(self),
                (other_unit for other_unit in other if other_unit.tag not in (self_unit.tag for self_unit in self)),
            ),
            self._bot_object,
        )

    def __add__(self, other: Units) -> Units:
        """
        :param other:
        """
        return Units(
            chain(
                iter(self),
                (other_unit for other_unit in other if other_unit.tag not in (self_unit.tag for self_unit in self)),
            ),
            self._bot_object,
        )

    def __and__(self, other: Units) -> Units:
        """
        :param other:
        """
        return Units(
            (other_unit for other_unit in other if other_unit.tag in (self_unit.tag for self_unit in self)),
            self._bot_object,
        )

    def __sub__(self, other: Units) -> Units:
        """
        :param other:
        """
        return Units(
            (self_unit for self_unit in self if self_unit.tag not in (other_unit.tag for other_unit in other)),
            self._bot_object,
        )

    def __hash__(self) -> int:
        return hash(unit.tag for unit in self)

    @property
    def amount(self) -> int:
        return len(self)

    @property
    def empty(self) -> bool:
        return not bool(self)

    @property
    def exists(self) -> bool:
        return bool(self)

    def find_by_tag(self, tag: int) -> Unit | None:
        """
        :param tag:
        """
        for unit in self:
            if unit.tag == tag:
                return unit
        return None

    def by_tag(self, tag: int) -> Unit:
        """
        :param tag:
        """
        unit = self.find_by_tag(tag)
        if unit is None:
            raise KeyError("Unit not found")
        return unit

    @property
    def first(self) -> Unit:
        assert self, "Units object is empty"
        return self[0]

    def take(self, n: int) -> Units:
        """
        :param n:
        """
        if n >= self.amount:
            return self
        return self.subgroup(self[:n])

    @property
    def random(self) -> Unit:
        assert self, "Units object is empty"
        return random.choice(self)

    def random_or(self, other: Any) -> Unit:
        return random.choice(self) if self else other

    def random_group_of(self, n: int) -> Units:
        """Returns self if n >= self.amount."""
        if n < 1:
            return Units([], self._bot_object)
        if n >= self.amount:
            return self
        return self.subgroup(random.sample(self, n))

    def in_attack_range_of(self, unit: Unit, bonus_distance: float = 0) -> Units:
        """Filters units that are in attack range of the given unit.
        This uses the unit and target unit.radius when calculating the distance, so it should be accurate.
        Caution: This may not work well for static structures (bunker, sieged tank, planetary fortress, photon cannon, spine and spore crawler) because it seems attack ranges differ for static / immovable units.

        Example::

            enemy_zerglings = self.enemy_units(UnitTypeId.ZERGLING)
            my_marine = next((unit for unit in self.units if unit.type_id == UnitTypeId.MARINE), None)
            if my_marine:
                all_zerglings_my_marine_can_attack = enemy_zerglings.in_attack_range_of(my_marine)

        Example::

            enemy_mutalisks = self.enemy_units(UnitTypeId.MUTALISK)
            my_marauder = next((unit for unit in self.units if unit.type_id == UnitTypeId.MARAUDER), None)
            if my_marauder:
                all_mutalisks_my_marauder_can_attack = enemy_mutaliskss.in_attack_range_of(my_marauder)
                # Is empty because mutalisk are flying and marauder cannot attack air

        :param unit:
        :param bonus_distance:
        """
        return self.filter(lambda x: unit.target_in_range(x, bonus_distance=bonus_distance))

    def closest_distance_to(self, position: Unit | Point2) -> float:
        """Returns the distance between the closest unit from this group to the target unit.

        Example::

            enemy_zerglings = self.enemy_units(UnitTypeId.ZERGLING)
            my_marine = next((unit for unit in self.units if unit.type_id == UnitTypeId.MARINE), None)
            if my_marine:
                closest_zergling_distance = enemy_zerglings.closest_distance_to(my_marine)
            # Contains the distance between the marine and the closest zergling

        :param position:
        """
        assert self, "Units object is empty"
        if isinstance(position, Unit):
            return min(self._bot_object._distance_squared_unit_to_unit(unit, position) for unit in self) ** 0.5
        return min(self._bot_object._distance_units_to_pos(self, position))

    def furthest_distance_to(self, position: Unit | Point2) -> float:
        """Returns the distance between the furthest unit from this group to the target unit


        Example::

            enemy_zerglings = self.enemy_units(UnitTypeId.ZERGLING)
            my_marine = next((unit for unit in self.units if unit.type_id == UnitTypeId.MARINE), None)
            if my_marine:
                furthest_zergling_distance = enemy_zerglings.furthest_distance_to(my_marine)
                # Contains the distance between the marine and the furthest away zergling

        :param position:
        """
        assert self, "Units object is empty"
        if isinstance(position, Unit):
            return max(self._bot_object._distance_squared_unit_to_unit(unit, position) for unit in self) ** 0.5
        return max(self._bot_object._distance_units_to_pos(self, position))

    def closest_to(self, position: Unit | Point2) -> Unit:
        """Returns the closest unit (from this Units object) to the target unit or position.

        Example::

            enemy_zerglings = self.enemy_units(UnitTypeId.ZERGLING)
            my_marine = next((unit for unit in self.units if unit.type_id == UnitTypeId.MARINE), None)
            if my_marine:
                closest_zergling = enemy_zerglings.closest_to(my_marine)
                # Contains the zergling that is closest to the target marine

        :param position:
        """
        assert self, "Units object is empty"
        if isinstance(position, Unit):
            return min(
                (unit1 for unit1 in self),
                key=lambda unit2: self._bot_object._distance_squared_unit_to_unit(unit2, position),
            )

        distances = self._bot_object._distance_units_to_pos(self, position)
        return min(((unit, dist) for unit, dist in zip(self, distances)), key=lambda my_tuple: my_tuple[1])[0]

    def furthest_to(self, position: Unit | Point2) -> Unit:
        """Returns the furhest unit (from this Units object) to the target unit or position.

        Example::

            enemy_zerglings = self.enemy_units(UnitTypeId.ZERGLING)
            my_marine = next((unit for unit in self.units if unit.type_id == UnitTypeId.MARINE), None)
            if my_marine:
                furthest_zergling = enemy_zerglings.furthest_to(my_marine)
                # Contains the zergling that is furthest away to the target marine

        :param position:
        """
        assert self, "Units object is empty"
        if isinstance(position, Unit):
            return max(
                (unit1 for unit1 in self),
                key=lambda unit2: self._bot_object._distance_squared_unit_to_unit(unit2, position),
            )
        distances = self._bot_object._distance_units_to_pos(self, position)
        return max(((unit, dist) for unit, dist in zip(self, distances)), key=lambda my_tuple: my_tuple[1])[0]

    def closer_than(self, distance: float, position: Unit | Point2) -> Units:
        """Returns all units (from this Units object) that are closer than 'distance' away from target unit or position.

        Example::

            enemy_zerglings = self.enemy_units(UnitTypeId.ZERGLING)
            my_marine = next((unit for unit in self.units if unit.type_id == UnitTypeId.MARINE), None)
            if my_marine:
                close_zerglings = enemy_zerglings.closer_than(3, my_marine)
                # Contains all zerglings that are distance 3 or less away from the marine (does not include unit radius in calculation)

        :param distance:
        :param position:
        """
        if not self:
            return self
        if isinstance(position, Unit):
            distance_squared = distance**2
            return self.subgroup(
                unit
                for unit in self
                if self._bot_object._distance_squared_unit_to_unit(unit, position) < distance_squared
            )
        distances = self._bot_object._distance_units_to_pos(self, position)
        return self.subgroup(unit for unit, dist in zip(self, distances) if dist < distance)

    def further_than(self, distance: float, position: Unit | Point2) -> Units:
        """Returns all units (from this Units object) that are further than 'distance' away from target unit or position.

        Example::

            enemy_zerglings = self.enemy_units(UnitTypeId.ZERGLING)
            my_marine = next((unit for unit in self.units if unit.type_id == UnitTypeId.MARINE), None)
            if my_marine:
                far_zerglings = enemy_zerglings.further_than(3, my_marine)
                # Contains all zerglings that are distance 3 or more away from the marine (does not include unit radius in calculation)

        :param distance:
        :param position:
        """
        if not self:
            return self
        if isinstance(position, Unit):
            distance_squared = distance**2
            return self.subgroup(
                unit
                for unit in self
                if distance_squared < self._bot_object._distance_squared_unit_to_unit(unit, position)
            )
        distances = self._bot_object._distance_units_to_pos(self, position)
        return self.subgroup(unit for unit, dist in zip(self, distances) if distance < dist)

    def in_distance_between(
        self, position: Unit | Point2 | tuple[float, float], distance1: float, distance2: float
    ) -> Units:
        """Returns units that are further than distance1 and closer than distance2 to unit or position.

        Example::

            enemy_zerglings = self.enemy_units(UnitTypeId.ZERGLING)
            my_marine = next((unit for unit in self.units if unit.type_id == UnitTypeId.MARINE), None)
            if my_marine:
                zerglings_filtered = enemy_zerglings.in_distance_between(my_marine, 3, 5)
                # Contains all zerglings that are between distance 3 and 5 away from the marine (does not include unit radius in calculation)

        :param position:
        :param distance1:
        :param distance2:
        """
        if not self:
            return self
        if isinstance(position, Unit):
            distance1_squared = distance1**2
            distance2_squared = distance2**2
            return self.subgroup(
                unit
                for unit in self
                if distance1_squared
                < self._bot_object._distance_squared_unit_to_unit(unit, position)
                < distance2_squared
            )
        distances = self._bot_object._distance_units_to_pos(self, position)
        return self.subgroup(unit for unit, dist in zip(self, distances) if distance1 < dist < distance2)

    def closest_n_units(self, position: Unit | Point2, n: int) -> Units:
        """Returns the n closest units in distance to position.

        Example::

            enemy_zerglings = self.enemy_units(UnitTypeId.ZERGLING)
            my_marine = next((unit for unit in self.units if unit.type_id == UnitTypeId.MARINE), None)
            if my_marine:
                zerglings_filtered = enemy_zerglings.closest_n_units(my_marine, 5)
                # Contains 5 zerglings that are the closest to the marine

        :param position:
        :param n:
        """
        if not self:
            return self
        return self.subgroup(self._list_sorted_by_distance_to(position)[:n])

    def furthest_n_units(self, position: Unit | Point2, n: int) -> Units:
        """Returns the n furhest units in distance to position.

        Example::

            enemy_zerglings = self.enemy_units(UnitTypeId.ZERGLING)
            my_marine = next((unit for unit in self.units if unit.type_id == UnitTypeId.MARINE), None)
            if my_marine:
                zerglings_filtered = enemy_zerglings.furthest_n_units(my_marine, 5)
                # Contains 5 zerglings that are the furthest to the marine

        :param position:
        :param n:
        """
        if not self:
            return self
        return self.subgroup(self._list_sorted_by_distance_to(position)[-n:])

    def in_distance_of_group(self, other_units: Units, distance: float) -> Units:
        """Returns units that are closer than distance from any unit in the other units object.

        :param other_units:
        :param distance:
        """
        assert other_units, "Other units object is empty"
        # Return self because there are no enemies
        if not self:
            return self
        distance_squared = distance**2
        if len(self) == 1:
            if any(
                self._bot_object._distance_squared_unit_to_unit(self[0], target) < distance_squared
                for target in other_units
            ):
                return self
            return self.subgroup([])

        return self.subgroup(
            self_unit
            for self_unit in self
            if any(
                self._bot_object._distance_squared_unit_to_unit(self_unit, other_unit) < distance_squared
                for other_unit in other_units
            )
        )

    def in_closest_distance_to_group(self, other_units: Units) -> Unit:
        """Returns unit in shortest distance from any unit in self to any unit in group.

        Loops over all units in self, then loops over all units in other_units and calculates the shortest distance. Returns the units that is closest to any unit of 'other_units'.

        :param other_units:
        """
        assert self, "Units object is empty"
        assert other_units, "Given units object is empty"
        return min(
            self,
            key=lambda self_unit: min(
                self._bot_object._distance_squared_unit_to_unit(self_unit, other_unit) for other_unit in other_units
            ),
        )

    def _list_sorted_closest_to_distance(self, position: Unit | Point2, distance: float) -> list[Unit]:
        """This function should be a bit faster than using units.sorted(key=lambda u: u.distance_to(position))

        :param position:
        :param distance:
        """
        if isinstance(position, Unit):
            return sorted(
                self,
                key=lambda unit: abs(self._bot_object._distance_squared_unit_to_unit(unit, position) - distance),
                reverse=True,
            )
        distances = self._bot_object._distance_units_to_pos(self, position)
        unit_dist_dict = {unit.tag: dist for unit, dist in zip(self, distances)}
        return sorted(self, key=lambda unit2: abs(unit_dist_dict[unit2.tag] - distance), reverse=True)

    def n_closest_to_distance(self, position: Point2, distance: float, n: int) -> Units:
        """Returns n units that are the closest to distance away.
        For example if the distance is set to 5 and you want 3 units, from units with distance [3, 4, 5, 6, 7] to position,
        the units with distance [4, 5, 6] will be returned

        :param position:
        :param distance:
        """
        return self.subgroup(self._list_sorted_closest_to_distance(position=position, distance=distance)[:n])

    def n_furthest_to_distance(self, position: Point2, distance: float, n: int) -> Units:
        """Inverse of the function 'n_closest_to_distance', returns the furthest units instead

        :param position:
        :param distance:
        """
        return self.subgroup(self._list_sorted_closest_to_distance(position=position, distance=distance)[-n:])

    def subgroup(self, units: Iterable[Unit]) -> Units:
        """Creates a new mutable Units object from Units or list object.

        :param units:
        """
        return Units(units, self._bot_object)

    def filter(self, pred: Callable[[Unit], Any]) -> Units:
        """Filters the current Units object and returns a new Units object.

        Example::

            from sc2.ids.unit_typeid import UnitTypeId
            my_marines = self.units.filter(lambda unit: unit.type_id == UnitTypeId.MARINE)

            completed_structures = self.structures.filter(lambda structure: structure.is_ready)

            queens_with_energy_to_inject = self.units.filter(lambda unit: unit.type_id == UnitTypeId.QUEEN and unit.energy >= 25)

            orbitals_with_energy_to_mule = self.structures.filter(lambda structure: structure.type_id == UnitTypeId.ORBITALCOMMAND and structure.energy >= 50)

            my_units_that_can_shoot_up = self.units.filter(lambda unit: unit.can_attack_air)

        See more unit properties in unit.py

        :param pred:
        """
        assert callable(pred), "Function is not callable"
        return self.subgroup(filter(pred, self))

    def sorted(self, key: Callable[[Unit], Any], reverse: bool = False) -> Units:
        return self.subgroup(sorted(self, key=key, reverse=reverse))

    def _list_sorted_by_distance_to(self, position: Unit | Point2, reverse: bool = False) -> list[Unit]:
        """This function should be a bit faster than using units.sorted(key=lambda u: u.distance_to(position))

        :param position:
        :param reverse:
        """
        if isinstance(position, Unit):
            return sorted(
                self, key=lambda unit: self._bot_object._distance_squared_unit_to_unit(unit, position), reverse=reverse
            )
        distances = self._bot_object._distance_units_to_pos(self, position)
        unit_dist_dict = {unit.tag: dist for unit, dist in zip(self, distances)}
        return sorted(self, key=lambda unit2: unit_dist_dict[unit2.tag], reverse=reverse)

    def sorted_by_distance_to(self, position: Unit | Point2, reverse: bool = False) -> Units:
        """This function should be a bit faster than using units.sorted(key=lambda u: u.distance_to(position))

        :param position:
        :param reverse:
        """
        return self.subgroup(self._list_sorted_by_distance_to(position, reverse=reverse))

    def tags_in(self, other: Iterable[int]) -> Units:
        """Filters all units that have their tags in the 'other' set/list/dict

        Example::

            my_inject_queens = self.units.tags_in(self.queen_tags_assigned_to_do_injects)

            # Do not use the following as it is slower because it first loops over all units to filter out if they are queens and loops over those again to check if their tags are in the list/set
            my_inject_queens_slow = self.units(QUEEN).tags_in(self.queen_tags_assigned_to_do_injects)

        :param other:
        """
        return self.filter(lambda unit: unit.tag in other)

    def tags_not_in(self, other: Iterable[int]) -> Units:
        """Filters all units that have their tags not in the 'other' set/list/dict

        Example::

            my_non_inject_queens = self.units.tags_not_in(self.queen_tags_assigned_to_do_injects)

            # Do not use the following as it is slower because it first loops over all units to filter out if they are queens and loops over those again to check if their tags are in the list/set
            my_non_inject_queens_slow = self.units(QUEEN).tags_not_in(self.queen_tags_assigned_to_do_injects)

        :param other:
        """
        return self.filter(lambda unit: unit.tag not in other)

    def of_type(self, other: UnitTypeId | Iterable[UnitTypeId]) -> Units:
        """Filters all units that are of a specific type

        Example::

            # Use a set instead of lists in the argument
            some_attack_units = self.units.of_type({ZERGLING, ROACH, HYDRALISK, BROODLORD})

        :param other:
        """
        if isinstance(other, UnitTypeId):
            other = {other}
        elif isinstance(other, list):
            other = set(other)
        return self.filter(lambda unit: unit.type_id in other)

    def exclude_type(self, other: UnitTypeId | Iterable[UnitTypeId]) -> Units:
        """Filters all units that are not of a specific type

        Example::

            # Use a set instead of lists in the argument
            ignore_units = self.enemy_units.exclude_type({LARVA, EGG, OVERLORD})

        :param other:
        """
        if isinstance(other, UnitTypeId):
            other = {other}
        elif isinstance(other, list):
            other = set(other)
        return self.filter(lambda unit: unit.type_id not in other)

    def same_tech(self, other: set[UnitTypeId]) -> Units:
        """Returns all structures that have the same base structure.

        Untested: This should return the equivalents for WarpPrism, Observer, Overseer, SupplyDepot and others

        Example::

            # All command centers, flying command centers, orbital commands, flying orbital commands, planetary fortress
            terran_townhalls = self.townhalls.same_tech(UnitTypeId.COMMANDCENTER)

            # All hatcheries, lairs and hives
            zerg_townhalls = self.townhalls.same_tech({UnitTypeId.HATCHERY})

            # All spires and greater spires
            spires = self.townhalls.same_tech({UnitTypeId.SPIRE})
            # The following returns the same
            spires = self.townhalls.same_tech({UnitTypeId.GREATERSPIRE})

            # This also works with multiple unit types
            zerg_townhalls_and_spires = self.structures.same_tech({UnitTypeId.HATCHERY, UnitTypeId.SPIRE})

        :param other:
        """
        assert isinstance(other, set), (
            "Please use a set as this filter function is already fairly slow. For example"
            + " 'self.units.same_tech({UnitTypeId.LAIR})'"
        )
        tech_alias_types: set[int] = {u.value for u in other}
        unit_data = self._bot_object.game_data.units
        for unit_type in other:
            for same in unit_data[unit_type.value]._proto.tech_alias:
                tech_alias_types.add(same)
        return self.filter(
            lambda unit: unit._proto.unit_type in tech_alias_types
            or any(same in tech_alias_types for same in unit._type_data._proto.tech_alias)
        )

    def same_unit(self, other: UnitTypeId | Iterable[UnitTypeId]) -> Units:
        """Returns all units that have the same base unit while being in different modes.

        Untested: This should return the equivalents for WarpPrism, Observer, Overseer, SupplyDepot and other units that have different modes but still act as the same unit

        Example::

            # All command centers on the ground and flying
            ccs = self.townhalls.same_unit(UnitTypeId.COMMANDCENTER)

            # All orbital commands on the ground and flying
            ocs = self.townhalls.same_unit(UnitTypeId.ORBITALCOMMAND)

            # All roaches and burrowed roaches
            roaches = self.units.same_unit(UnitTypeId.ROACH)
            # This is useful because roach has a different type id when burrowed
            burrowed_roaches = self.units(UnitTypeId.ROACHBURROWED)

        :param other:
        """
        if isinstance(other, UnitTypeId):
            other = {other}
        unit_alias_types: set[int] = {u.value for u in other}
        unit_data = self._bot_object.game_data.units
        for unit_type in other:
            unit_alias_types.add(unit_data[unit_type.value]._proto.unit_alias)
        unit_alias_types.discard(0)
        return self.filter(
            lambda unit: unit._proto.unit_type in unit_alias_types
            or unit._type_data._proto.unit_alias in unit_alias_types
        )

    @property
    def center(self) -> Point2:
        """Returns the central position of all units."""
        assert self, "Units object is empty"
        return Point2(
            (
                sum(unit._proto.pos.x for unit in self) / self.amount,
                sum(unit._proto.pos.y for unit in self) / self.amount,
            )
        )

    @property
    def selected(self) -> Units:
        """Returns all units that are selected by the human player."""
        return self.filter(lambda unit: unit.is_selected)

    @property
    def tags(self) -> set[int]:
        """Returns all unit tags as a set."""
        return {unit.tag for unit in self}

    @property
    def ready(self) -> Units:
        """Returns all structures that are ready (construction complete)."""
        return self.filter(lambda unit: unit.is_ready)

    @property
    def not_ready(self) -> Units:
        """Returns all structures that are not ready (construction not complete)."""
        return self.filter(lambda unit: not unit.is_ready)

    @property
    def idle(self) -> Units:
        """Returns all units or structures that are doing nothing (unit is standing still, structure is doing nothing)."""
        return self.filter(lambda unit: unit.is_idle)

    @property
    def owned(self) -> Units:
        """Deprecated: All your units."""
        return self.filter(lambda unit: unit.is_mine)

    @property
    def enemy(self) -> Units:
        """Deprecated: All enemy units."""
        return self.filter(lambda unit: unit.is_enemy)

    @property
    def flying(self) -> Units:
        """Returns all units that are flying."""
        return self.filter(lambda unit: unit.is_flying)

    @property
    def not_flying(self) -> Units:
        """Returns all units that not are flying."""
        return self.filter(lambda unit: not unit.is_flying)

    @property
    def structure(self) -> Units:
        """Deprecated: All structures."""
        return self.filter(lambda unit: unit.is_structure)

    @property
    def not_structure(self) -> Units:
        """Deprecated: All units that are not structures."""
        return self.filter(lambda unit: not unit.is_structure)

    @property
    def gathering(self) -> Units:
        """Returns all workers that are mining minerals or vespene (gather command)."""
        return self.filter(lambda unit: unit.is_gathering)

    @property
    def returning(self) -> Units:
        """Returns all workers that are carrying minerals or vespene and are returning to a townhall."""
        return self.filter(lambda unit: unit.is_returning)

    @property
    def collecting(self) -> Units:
        """Returns all workers that are mining or returning resources."""
        return self.filter(lambda unit: unit.is_collecting)

    @property
    def visible(self) -> Units:
        """Returns all units or structures that are visible.
        TODO: add proper description on which units are exactly visible (not snapshots?)"""
        return self.filter(lambda unit: unit.is_visible)

    @property
    def mineral_field(self) -> Units:
        """Returns all units that are mineral fields."""
        return self.filter(lambda unit: unit.is_mineral_field)

    @property
    def vespene_geyser(self) -> Units:
        """Returns all units that are vespene geysers."""
        return self.filter(lambda unit: unit.is_vespene_geyser)

    @property
    def prefer_idle(self) -> Units:
        """Sorts units based on if they are idle. Idle units come first."""
        return self.sorted(lambda unit: unit.is_idle, reverse=True)
